//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "main.h"
#include "util.h"
#include "tileunit.h"
#include "mapunit.h"
#include "about.h"
#include "xFiles.hpp"

//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TMainForm *MainForm;
//---------------------------------------------------------------------------
__fastcall TMainForm::TMainForm(TComponent* Owner)
        : TForm(Owner)
{
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::UpdateView()
{
  DrawBackBitmap();
  pbxView->Canvas->Draw(0, 0, FBackBitmap);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::pbxViewPaint(TObject *Sender)
{
  UpdateView();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FormCreate(TObject *Sender)
{
  randomize(); // üƺؤl

  try {
    FRole.LoadBits(); // ŪJDϧ

    Terrs.LoadFromFile(AppDir + FN_TERR_ARCHIVE); // ŪJa
    Items.LoadFromFile(AppDir + FN_ITEM_ARCHIVE); // ŪJ~
  } catch (Exception& E) {
    ShowMessage(E.Message);
    PostMessage(Handle, WM_CLOSE, 0, 0); // H asynchronous 覡
  }

  // According tile num and size, adjust dimension of pnlView
  pnlView->ClientWidth = TILE_WIDTH * TILE_NUM_X;
  pnlView->ClientHeight = TILE_HEIGHT * TILE_NUM_Y;

  ClientWidth = pnlView->Width;
  ClientHeight = pnlView->Height;

  FBackBitmap = new Graphics::TBitmap; // wĥ bitmap
  FBackBitmap->Width = TILE_WIDTH * TILE_NUM_X;
  FBackBitmap->Height = TILE_HEIGHT * TILE_NUM_Y;

  FPlayBackLevelNo = 0; // ثe PlayBack
  FLevelNo = 1;

  // wƤu@nF, iJ title e
  GameStatus = gsTitle;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::FormClose(TObject *Sender, TCloseAction &Action)
{
  delete FBackBitmap;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::DrawBackBitmap()
{
  Map.DrawTerrMap(FBackBitmap->Canvas); // øsaμh
  Map.DrawItemMap(FBackBitmap->Canvas); // øs~h
  FRole.Draw(FBackBitmap->Canvas); // eXϮ

  int M = FPlayingTime / 60; // wOɶ
  int S = FPlayingTime % 60; // wOɶ

  FBackBitmap->Canvas->Font->Color = clWhite;
  FBackBitmap->Canvas->Font->Name = "FixedSys";
  FBackBitmap->Canvas->Font->Style = TFontStyles() << fsBold;
  FBackBitmap->Canvas->Font->Size = 14;
  FBackBitmap->Canvas->Brush->Style = bsClear;

  TRect R;
  switch (FGameStatus) {
    case gsTitle:
      // eXWDjrΤU誺@̦W
      R = Rect(0, 0, TILE_WIDTH * TILE_NUM_X, TILE_HEIGHT * TILE_NUM_Y - FBackBitmap->Canvas->TextHeight("") / 2);
      DrawText(FBackBitmap->Canvas->Handle, "@: eF", - 1, &R, DT_BOTTOM | DT_CENTER | DT_SINGLELINE);
      DrawStatusBox("w{ yf", true);
      break;

    case gsPlayBack: FBackBitmap->Canvas->TextOut(5, 5, Format("L^U (LEVEL %d) ...", OPENARRAY(TVarRec, (FLevelNo))));
      break;

    default:
      // bkWܮɶ, H XX:XX 榡
      AnsiString Str = Format("%.2d:%.2d", OPENARRAY(TVarRec, (M, S)));
      FBackBitmap->Canvas->TextOut(TILE_WIDTH * TILE_NUM_X - FBackBitmap->Canvas->TextWidth(Str) - 5, 5, Str);

      // bWd
      Str = Format("LEVEL %d", OPENARRAY(TVarRec, (FLevelNo)));
      FBackBitmap->Canvas->TextOut(5, 5, Str);
      break;
  }

  // oOLe
  if (FGameStatus == gsSuccess)
    DrawStatusBox("z, \F !!", false);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::SetGameStatus(TGameStatus Value)
{
  FGameStatus = Value;

  // ھڷsCA}Tӭpɾ
  tmrTimer->Enabled = FGameStatus == gsPlaying;
  tmrPlayBack->Enabled = FGameStatus == gsPlayBack;
  tmrTitle->Enabled = FGameStatus == gsTitle;

  switch (FGameStatus) {
    case gsTitle: LevelNo = 1; // weܲĤ@a
      break;

    case gsPlaying:
      FPlayingTime = 0; // pks
      FRole.CleanMoveList(); // play back ks
      break;

    case gsSuccess:
      // LF, NdO_, ܭnɴN^d
      FPlayBackLevelNo = FLevelNo;
      break;

    case gsPlayBack:
      LevelNo = FPlayBackLevelNo; // O^Ud
      tmrPlayBack->Tag = 0; // qĤ@B}l
      break;
  }

  UpdateControlStatus(); // sDCΨ䥦
  UpdateView(); // sCe
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::SetLevelNo(int Value)
{
  Map.LevelNo = Value;
  FRole.X = Map.Role_X;
  FRole.Y = Map.Role_Y;

  FLevelNo = Value;
  UpdateControlStatus();
  UpdateView();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::UpdateControlStatus()
{
  const AnsiString StatusStr[4] = {"Welcome", "", "Congratulation !!", "PlayBack"};

  if (FGameStatus != gsPlaying)
    Caption = Format("%s - %s", OPENARRAY(TVarRec, (Application->Title, StatusStr[(int)FGameStatus])));
  else
    Caption = Format("%s - Level %d", OPENARRAY(TVarRec, (Application->Title, FLevelNo)));

  mnuRestart->Enabled = FGameStatus == gsPlaying;
  mnuStartGame->Enabled = FGameStatus != gsPlaying;

  mnuPrevLevel->Enabled = FGameStatus == gsPlaying && LevelNo > 1;
  mnuNextLevel->Enabled = FGameStatus == gsPlaying;

  TDirectionArray& PlayBackList = FRole.PlayBackList;
  mnuPlayBack->Enabled = (FPlayBackLevelNo != 0) && (PlayBackList.size() > 0);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::DrawStatusBox(AnsiString S, bool TopOrCenter)
{
  FBackBitmap->Canvas->Font->Color = clWhite; // r
  FBackBitmap->Canvas->Font->Name = "FixedSys";
  FBackBitmap->Canvas->Font->Style = TFontStyles() << fsBold;
  FBackBitmap->Canvas->Font->Size = 20;

  TSize TextSize = FBackBitmap->Canvas->TextExtent(S);
  // first calculate the necessary rectangle
  TRect R = Rect(0, 0, TextSize.cx, TextSize.cy);
  OffsetRect(&R, (FBackBitmap->Width - TextSize.cx) / 2, (FBackBitmap->Height - TextSize.cy) / 2); // Nخظme

  if (TopOrCenter) {
    R.Top = 20; // Nخ\bW褤
    R.Bottom = R.Top + TextSize.cy;
  }

  InflateRect(&R, 10, 10); // yø~, e׬ 10
  FBackBitmap->Canvas->Pen->Width = 4;
  FBackBitmap->Canvas->Pen->Color = clWhite;
  FBackBitmap->Canvas->Pen->Style = psSolid;
  FBackBitmap->Canvas->Brush->Color = clBlue;
  FBackBitmap->Canvas->Brush->Style = bsSolid;
  FBackBitmap->Canvas->Rectangle(R.Left, R.Top, R.Right, R.Bottom);

  // NreX, m
  DrawText(FBackBitmap->Canvas->Handle, S.c_str(), S.Length(), &R, DT_CENTER | DT_SINGLELINE | DT_VCENTER);
}
//---------------------------------------------------------------------------


void __fastcall TMainForm::mnuStartGameClick(TObject *Sender)
{
  GameStatus = gsPlaying;
  LevelNo = 1;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::mnuChooseLevelClick(TObject *Sender)
{
  AnsiString S = IntToStr(FLevelNo);
  if (InputQuery(Application->Title, "Input Level No:", S)) {
    GameStatus = gsPlaying;
    LevelNo = StrToInt(S);
  }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::mnuRestartClick(TObject *Sender)
{
  LevelNo = LevelNo;        
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::mnuPrevLevelClick(TObject *Sender)
{
  switch (dynamic_cast<TComponent*>(Sender)->Tag) {
    case 0: LevelNo = LevelNo - 1; break;
    case 1: LevelNo = LevelNo + 1; break;
  }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::mnuExitClick(TObject *Sender)
{
  Close();        
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::mnuPlayBackClick(TObject *Sender)
{
  GameStatus = gsPlayBack;        
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::mnuAboutClick(TObject *Sender)
{
  TAboutBox* box = new TAboutBox(this);
  try {
    box->ShowModal();
  } __finally {
    delete box;
  }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::tmrTimerTimer(TObject *Sender)
{
  FPlayingTime++; // Ciɶ[@
  UpdateView(); // jse
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::tmrPlayBackTimer(TObject *Sender)
{
  // Ӥeʧ@`Ǩ, ϥ tmrPlayBack  Tag ݩʨӰOثeĴXB
  TDirectionArray& plist = FRole.PlayBackList;
  FRole.Move(plist[tmrPlayBack->Tag]);

  tmrPlayBack->Tag = tmrPlayBack->Tag + 1; // o@BLF, WܤU@B

  if ((unsigned int)tmrPlayBack->Tag == plist.size()) // F, Le
    GameStatus = gsSuccess;

  UpdateView(); // se
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::tmrTitleTimer(TObject *Sender)
{
  FRole.Move((TDirection)random(4)); // DN
  UpdateView(); // se
}
//---------------------------------------------------------------------------
bool __fastcall TMainForm::CheckFinished()
{
  // v@ˬdCӥتaάO_wQл\Ϊ~л\F ?
  for (int y = 0; y < TILE_NUM_Y; y++)
    for (int x = 0; x < TILE_NUM_X; x++)
      if (Map.IsTarget[x][y] && !Map.IsSource[x][y])
        return false; // @, @ӥتa٨SQ,  !!

  return true; // Yeah, ȹF
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::FormKeyDown(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
  if (FGameStatus == gsPlaying) { // `N, uCA, L~
    switch (Key) {
      case VK_UP: FRole.Move(drUp); break; // VW

      case VK_DOWN: FRole.Move(drDown); break; // VU

      case VK_LEFT: FRole.Move(drLeft); break; // V

      case VK_RIGHT: FRole.Move(drRight); break; // Vk

      default: return; // ë@q, z
    }

    UpdateView(); // se

    if (CheckFinished()) { // O_ ??
      FRole.SavePlayBack(); // NʰOs_ӥHK
      GameStatus = gsSuccess; // iJLA
    }
  }
}
//---------------------------------------------------------------------------

