//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "main.h"

#include "util.h"
#include "tileunit.h"
#include "mapunit.h"
#include "xFiles.hpp"
#include "xUtils.hpp"
#include "about.h"

//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TMainForm *MainForm;

//---------------------------------------------------------------------------
__fastcall TMainForm::TMainForm(TComponent* Owner)
        : TForm(Owner)
{
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FormCreate(TObject *Sender)
{
  FEditKind = ekTerrs; // w]aνsҦ
  FButtonPressed = mbMiddle; // ܥثeƹS
  
  try {
    FRole.LoadBits(); // ŪJDϧ

    Terrs.LoadFromFile(AppDir + FN_TERR_ARCHIVE); // ŪJa
    Items.LoadFromFile(AppDir + FN_ITEM_ARCHIVE); // ŪJ~
  } catch (Exception& E) {
    ShowMessage(E.Message);
    PostMessage(Handle, WM_CLOSE, 0, 0); // H asynchronous 覡
  }
  
  // According tile num and size, adjust dimension of pnlView
  pnlView->ClientWidth = TILE_WIDTH * TILE_NUM_X;
  pnlView->ClientHeight = TILE_HEIGHT * TILE_NUM_Y;
  
  FBackBitmap = new Graphics::TBitmap; // wĥ bitmap
  FBackBitmap->Width = TILE_WIDTH * TILE_NUM_X;
  FBackBitmap->Height = TILE_HEIGHT * TILE_NUM_Y;
  
  LevelNo = 1;
  UpdateControlStatus();
  
  // ƹЦm
  FCursorX = -1;
  FCursorY = -1;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::FormClose(TObject *Sender, TCloseAction &Action)
{
  delete FBackBitmap;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::pbxViewPaint(TObject *Sender)
{
  UpdateView();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::mnuExitClick(TObject *Sender)
{
  Close();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::mnuChooseLevelClick(TObject *Sender)
{
  if (!AskSaveMap()) return;

  AnsiString S = IntToStr(FLevelNo);
  if (InputQuery(Application->Title, "Input Level No:", S))
    LevelNo = StrToInt(S);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::mnuSaveClick(TObject *Sender)
{
  MySaveMap();        
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::mnuRestoreLevelClick(TObject *Sender)
{
  // |Ĳo property write method (SetLevelNo)
  LevelNo = LevelNo;        
}
//---------------------------------------------------------------------------

#define MAX(a, b) ((a)>(b)?(a):(b))

void __fastcall TMainForm::mnuEditKindClick(TObject *Sender)
{
  const TEditKind Kinds[3] = {ekTerrs, ekItems, ekRole};
  const TColor Colors[2] = {clBtnFace, clWindow};

  TMenuItem* item = dynamic_cast<TMenuItem*>(Sender);
  
  item->Checked = true;
  FEditKind = Kinds[item->Tag]; // ]wsҦ

  UpdateControlStatus();
  UpdateView();

  bool bEnable = FEditKind != ekRole; // Ys訤⪬A, hݥΨ grdPreview
  grdPreview->Enabled = bEnable;
  grdPreview->Color = Colors[bEnable ? 1 : 0];
  grdPreview->Invalidate();

  switch (FEditKind) {
    case ekTerrs:
      nbkTileAttr->PageIndex = 0;
      grdPreview->RowCount = MAX(10, Terrs.TileRowCount);
      break;

    case ekItems:
      nbkTileAttr->PageIndex = 1;
      grdPreview->RowCount = MAX(10, Items.TileRowCount);
      break;
  }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::mnuRemoveItemClick(TObject *Sender)
{
  if (FCursorX == -1 || FCursorY == -1) return;

  Map.ItemMap[FCursorX][FCursorY] = 0;
  UpdateView();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::mnuResetTerrsClick(TObject *Sender)
{
  switch (dynamic_cast<TComponent*>(Sender)->Tag) {
    case 0: Map.ResetTerrs(); break;
    case 1: Map.ResetItems(); break;
  }
  UpdateView();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::mnuShowSomethingClick(TObject *Sender)
{
  ((TMenuItem*)Sender)->Checked = !((TMenuItem*)Sender)->Checked;
  UpdateView();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::mnuAboutClick(TObject *Sender)
{
  TAboutBox* box = new TAboutBox(this);
  try {
    box->ShowModal();
  } __finally {
    delete box;
  }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::UpdateView()
{
  DrawBackBitmap();
  pbxView->Canvas->Draw(0, 0, FBackBitmap);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::DrawBackBitmap()
{
  Map.DrawTerrMap(FBackBitmap->Canvas); // øsaμh
  Map.DrawItemMap(FBackBitmap->Canvas); // øs~h

  // pGϥΪ̷Qݮu, NNueX
  if (mnuShowGrid->Checked) {
    FBackBitmap->Canvas->Pen->Color = clBlack;
    FBackBitmap->Canvas->Pen->Style = psSolid;
    FBackBitmap->Canvas->Pen->Width = 1;

    // eu
    for (int y = 0; y < TILE_NUM_Y; y++) {
      FBackBitmap->Canvas->MoveTo(0, y * TILE_HEIGHT);
      FBackBitmap->Canvas->LineTo(TILE_NUM_X * TILE_WIDTH, y * TILE_HEIGHT);
    }

    // Aeu
    for (int x = 0; x < TILE_NUM_X; x++) {
      FBackBitmap->Canvas->MoveTo(x * TILE_WIDTH, 0);
      FBackBitmap->Canvas->LineTo(x * TILE_WIDTH, TILE_NUM_Y * TILE_HEIGHT);
    }
  }

  // pGϥΪ̷QݯSϰ, NتaaμХX
  if (mnuShowSpeicalArea->Checked) {
    FBackBitmap->Canvas->Pen->Color = clRed;
    FBackBitmap->Canvas->Pen->Width = 1;
    FBackBitmap->Canvas->Pen->Style = psDash;
    FBackBitmap->Canvas->Brush->Style = bsClear;

    for (int y = 0; y < TILE_NUM_Y; y++) // v@ˬd
      for (int x = 0; x < TILE_NUM_X; x++)
        if (Map.IsTarget[x][y]) { // ~إ[WkWe쥪Uu
          FBackBitmap->Canvas->Rectangle(x * TILE_WIDTH, y * TILE_HEIGHT, (x + 1) * TILE_WIDTH - 1, (y + 1) * TILE_HEIGHT - 1);
          FBackBitmap->Canvas->MoveTo((x + 1) * TILE_WIDTH, y * TILE_HEIGHT);
          FBackBitmap->Canvas->LineTo(x * TILE_WIDTH, (y + 1) * TILE_HEIGHT);
        }
   }

  // pGϥΪ̷QݯS~, Nл\Ϊ~ХX
  if (mnuShowSpeicalItems->Checked) {
    FBackBitmap->Canvas->Pen->Color = clLime;
    FBackBitmap->Canvas->Pen->Width = 1;
    FBackBitmap->Canvas->Pen->Style = psDash;
    FBackBitmap->Canvas->Brush->Style = bsClear;

    for (int y = 0; y < TILE_NUM_Y; y++) // v@ˬd
      for (int x = 0; x < TILE_NUM_X; x++)
        if (Map.IsSource[x][y]) { // ~إ[WWekUGu
          FBackBitmap->Canvas->Rectangle(x * TILE_WIDTH, y * TILE_HEIGHT, (x + 1) * TILE_WIDTH - 1, (y + 1) * TILE_HEIGHT - 1);
          FBackBitmap->Canvas->MoveTo(x * TILE_WIDTH, y * TILE_HEIGHT);
          FBackBitmap->Canvas->LineTo((x + 1) * TILE_WIDTH, (y + 1) * TILE_HEIGHT);
        }
  }

  FRole.Draw(FBackBitmap->Canvas); // ̫eXϮ
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::UpdateControlStatus()
{
  Caption = Format("%s - Level %d", OPENARRAY(TVarRec, (Application->Title, FLevelNo)));
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::SetLevelNo(int Value)
{
  try {
    // ŪaɤΨm
    Map.LevelNo = Value;
  } catch(...) {
    // YŪaɥ, Mia
    Map.ResetMap();
  }
  
  FRole.X = Map.Role_X; // Nm Map 󤤧ۥX
  FRole.Y = Map.Role_Y;

  FLevelNo = Value;
  FModified = False; // aϩ|ʡ]M:p^
  UpdateControlStatus();
  UpdateView(); // OѤFsCe
}
//---------------------------------------------------------------------------
// ˬdaϬO_Xk
bool __fastcall TMainForm::ValidateMap()
{
  int x, y;
  int SourceCount, TargetCount;

  SourceCount = 0; // л\Ϊ~ƥ
  TargetCount = 0; // تaμƥ
  for (int y = 0; y < TILE_NUM_Y; y++)
    for (int x = 0; x < TILE_NUM_X; x++) {

      // pл\Ϊ~Υتaμƥ
      if (Map.IsSource[x][y]) SourceCount++;
      if (Map.IsTarget[x][y]) TargetCount++;

      // aϤOϤsOOjϮwϤƥ ? OܴNզ^w]
      if (Map.TerrMap[x][y] > Terrs.TileNum) Map.TerrMap[x][y] = 0;
      if (Map.ItemMap[x][y] > Items.TileNum) Map.ItemMap[x][y] = 0;
    }

  // O_Sتa ? (LkL)
  if (TargetCount == 0 &&
    !YesNoBox("Sتa, O_~H", MB_DEFBUTTON1)) return false;

  // O_л\Ϊ~֩تa ? (LkL)
  if (TargetCount > SourceCount &&
    !YesNoBox("л\Ϊ~֩تa, O_~H", MB_DEFBUTTON1)) return false;

  // 󤣯ಾʪaΤW
  if (!Map.CanPass[Map.Role_X][Map.Role_Y] &&
    !YesNoBox("D󤣯ಾʪaΤW, O_~H", MB_DEFBUTTON1)) return false;

  return true;
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------

void __fastcall TMainForm::pbxViewMouseMove(TObject *Sender,
      TShiftState Shift, int X, int Y)
{
  // wg]Xd~F...
  if (X < 0 ||  X >= TILE_WIDTH * TILE_NUM_X || Y < 0 ||
    Y >= TILE_HEIGHT * TILE_NUM_Y) return;

  FCursorX = X / TILE_WIDTH; // y, HϮ欰
  FCursorY = Y / TILE_HEIGHT;

  // @Ԧ@m񪫫~ĪG
  pbxView->OnMouseDown(Sender, FButtonPressed, Shift, X, Y);

  UpdateView(); // øCe

  // eXثeҿϰ~
  pbxView->Canvas->Pen->Width = 2;
  pbxView->Canvas->Pen->Color = clBlue;
  pbxView->Canvas->Brush->Style = bsClear;
  pbxView->Canvas->Rectangle(FCursorX * TILE_WIDTH, FCursorY * TILE_HEIGHT,
    (FCursorX + 1) * TILE_WIDTH, (FCursorY + 1) * TILE_HEIGHT);

  stbMain->SimpleText = Format("Position: (%d, %d)", OPENARRAY(TVarRec, (FCursorX, FCursorY)));
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::grdPreviewDrawCell(TObject *Sender, int ACol,
      int ARow, TRect &Rect, TGridDrawState State)
{
  bool bEmpty = false;  // O_OŪ, ݭn

  int No = ARow * grdPreview->ColCount + ACol;

  switch (FEditKind) {
    case ekTerrs:
      if (No < Terrs.TileNum)
        BitBlt(grdPreview->Canvas->Handle, Rect.Left, Rect.Top, TILE_WIDTH, TILE_HEIGHT,
        Terrs.Bitmap->Canvas->Handle, Terrs.TilePos_Left[No], Terrs.TilePos_Top[No], SRCCOPY);
      else
        bEmpty = true;
      break;

    case ekItems:
      if (No < Terrs.TileNum)
        BitBlt(grdPreview->Canvas->Handle, Rect.Left, Rect.Top, TILE_WIDTH, TILE_HEIGHT,
        Items.Bitmap->Canvas->Handle, Items.TilePos_Left[No], Items.TilePos_Top[No], SRCCOPY);
      else
        bEmpty = true;
      break;
  }

  if (bEmpty) {
    grdPreview->Canvas->Brush->Color = clBlack;
    grdPreview->Canvas->FillRect(Rect);
  }
}
//---------------------------------------------------------------------------
bool __fastcall TMainForm::AskSaveMap()
{
  if (!FModified) return true;

  // T{O_xsثes誺aɮ
  switch (YesNoCancelBox(Format("Do you want to save this map (level %d) ?", OPENARRAY(TVarRec, (FLevelNo))))) {
    case mrYes: MySaveMap(); break;

    case mrNo: break;

    case mrCancel: return false;
  }

  return true; // O_ cancel ʧ@
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::MySaveMap()
{
  if (!FModified) return;
  if (!ValidateMap()) return;

  Map.Role_X = FRole.X;
  Map.Role_Y = FRole.Y;
  Map.SaveToFile();
  FModified = false;
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------

void __fastcall TMainForm::grdPreviewClick(TObject *Sender)
{
  int No = grdPreview->Row * grdPreview->ColCount + grdPreview->Col;

  // ܿݩ
  switch (FEditKind) {
    case ekTerrs:
      cbxCanPass->Checked = Terrs.Attrs[No].Contains(taCanPass);
      cbxTarget->Checked = Terrs.Attrs[No].Contains(taTarget);
      break;

    case ekItems:
      cbxCanMove->Checked = Items.Attrs[No].Contains(iaCanMove);
      cbxSource->Checked = Items.Attrs[No].Contains(iaSource);
      break;
  }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::FormCloseQuery(TObject *Sender, bool &CanClose)
{
  if (!AskSaveMap()) CanClose = false;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::pbxViewMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
  FButtonPressed = mbMiddle; // ]w mbMiddle ܨϥΪ̤w}ƹ        
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::pbxViewMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
  // NUO_, tX OnMouseMove event handler ͩԦ]wĪG
  FButtonPressed = Button;

  if (Button == mbMiddle) return; // ƹ䤣

  int No = grdPreview->Row * grdPreview->ColCount + grdPreview->Col; // pثeܪϤs

  if (Button == mbLeft) { // O]w
    switch (FEditKind) { // ھڽsҦPi椣P]wʧ@
      case ekTerrs:
        Map.TerrMap[FCursorX][FCursorY] = No; // Nsa\W
        break;

      case ekItems:
        // ~i\b⨭W
        if (FRole.X == FCursorX && FRole.Y == FCursorY) return;

        Map.ItemMap[FCursorX][FCursorY] = No; // Ns\W
        break;

      case ekRole:
        // ⤣iH\b~W
        if (Map.ItemMap[FCursorX][FCursorY] != 0) return;
        // ⤣iH\biʪaΤW
        if (!Map.CanPass[FCursorX][FCursorY]) return;

        FRole.X = FCursorX; // ]wm
        FRole.Y = FCursorY;
        break;
    }
  } else { // kOM
    switch (FEditKind) {
      case ekTerrs: Map.TerrMap[FCursorX][FCursorY] = 0; // Ma
                    break;

      case ekItems: Map.ItemMap[FCursorX][FCursorY] = 0; // M~
                    break;

      case ekRole:   // ⤣M, so do nothing
                    break;
    }
  }

  FModified = true;
  UpdateView();
}
//---------------------------------------------------------------------------

