//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#define mapunitCPP
#include "mapunit.h"

#include <typeinfo>

#include "xFiles.hpp"
#pragma package(smart_init)
//---------------------------------------------------------------------------

TMap::TMap()
{
  FLevelNo = 0;

  FInvBitmap = new Graphics::TBitmap;
  FInvBitmap->Width = TILE_WIDTH;
  FInvBitmap->Height = TILE_HEIGHT;
  FInvBitmap->Transparent = true;
  FInvBitmap->TransparentColor = (TColor)RGB(128, 0, 0);
}

TMap::~TMap()
{
  delete FInvBitmap;
}

AnsiString TMap::GetFileName()
{
  return Format("%s%s%d%s", OPENARRAY(TVarRec, (AppDir, FN_MAP_PREFIX, FLevelNo, FN_MAP_EXT)));
}

void TMap::LoadFromFile()
{
  TFileStream* fs;
  TReader* reader;

  fs = new TFileStream(GetFileName(), fmOpenRead);
  reader = new TReader(fs, 2048);
  try {
    CheckSignature(reader, SIG_MYFILE);
    CheckSignature(reader, typeid(*this).name());

    FRole_X = reader->ReadInteger();
    FRole_Y = reader->ReadInteger();

    reader->Read(FTerrMap, sizeof(TMapArray));
    reader->Read(FItemMap, sizeof(TMapArray));
  } __finally {
    delete reader;
    delete fs;
  }
}

void TMap::SaveToFile()
{
  TFileStream* fs;
  TWriter* writer;

  fs = new TFileStream(GetFileName(), fmCreate | fmOpenWrite);
  writer = new TWriter(fs, 2048);
  try {
    writer->WriteString(SIG_MYFILE);
    writer->WriteString(typeid(*this).name());

    writer->WriteInteger(FRole_X);
    writer->WriteInteger(FRole_Y);

    writer->Write(FTerrMap, sizeof(TMapArray));
    writer->Write(FItemMap, sizeof(TMapArray));
  } __finally {
    delete writer;
    delete fs;
  }
}

void TMap::SetLevelNo(int Value)
{
  FLevelNo = Value;
  LoadFromFile();
}

bool TMap::GetCanPass(int x, int y)
{
  return Terrs.Attrs[FTerrMap[y][x]].Contains(taCanPass);
}

bool TMap::GetIsTarget(int x, int y)
{
  return Terrs.Attrs[FTerrMap[y][x]].Contains(taTarget);
}

TTerrAttr TMap::GetTerrAttr(int x, int y)
{
  return Terrs.Attrs[FTerrMap[y][x]];
}

void TMap::SetTerrAttr(int x, int y, TTerrAttr Value)
{
  Terrs.Attrs[FTerrMap[y][x]] = Value;
}

TItemAttr TMap::GetItemAttr(int x, int y)
{
  return Items.Attrs[FItemMap[y][x]];
}

void TMap::SetItemAttr(int x, int y, TItemAttr Value)
{
  Items.Attrs[FItemMap[y][x]] = Value;
}

bool TMap::GetCanMove(int x, int y)
{
  return Items.Attrs[FItemMap[y][x]].Contains(iaCanMove);
}

bool TMap::GetIsSource(int x, int y)
{
  return Items.Attrs[FItemMap[y][x]].Contains(iaSource);
}

void TMap::DrawTerrMap(TCanvas* Canvas)
{
  for (int y = 0; y < TILE_NUM_Y; y++)
    for (int x = 0; x < TILE_NUM_X; x++)
      BitBlt(Canvas->Handle, x * TILE_WIDTH, y * TILE_HEIGHT, TILE_WIDTH, TILE_HEIGHT,
      Terrs.Bitmap->Canvas->Handle, Terrs.TilePos_Left[FTerrMap[y][x]], Terrs.TilePos_Top[FTerrMap[y][x]], SRCCOPY);
}

void TMap::DrawItemMap(TCanvas* Canvas)
{
  // Τ@ FInvBitmap  Items.Bitmap 榡
  FInvBitmap->PixelFormat = Items.Bitmap->PixelFormat;
  
  for (int y = 0; y < TILE_NUM_Y; y++)
    for (int x = 0; x < TILE_NUM_X; x++)
      if (FItemMap[y][x]) {
        int Left = Items.TilePos_Left[FItemMap[y][x]];
        int Top = Items.TilePos_Top[FItemMap[y][x]];

        FInvBitmap->Canvas->CopyRect(Rect(0, 0, TILE_WIDTH, TILE_HEIGHT), Items.Bitmap->Canvas,
          Rect(Left, Top, Left + TILE_WIDTH, Top + TILE_HEIGHT));

        // ѨM VCL b Windows 98 UzK bug  workaround
        if (IsWindows98())
          DrawTransparentBitmap(Canvas->Handle, FInvBitmap->Handle, x * TILE_WIDTH, y * TILE_HEIGHT, RGB(128, 0, 0));
        else
          Canvas->Draw(x * TILE_WIDTH, y * TILE_HEIGHT, FInvBitmap);
      }
}

void TMap::ResetMap()
{
  ResetTerrs();
  ResetItems();
  
  FRole_X = 0;
  FRole_Y = 0;
}

void TMap::ResetTerrs()
{
  memset(FTerrMap, 0, sizeof(FTerrMap));
}

void TMap::ResetItems()
{
  memset(FItemMap, 0, sizeof(FItemMap));
}

Byte TMap::GetItemMap(int x, int y)
{
  return FItemMap[y][x];
}

Byte TMap::GetTerrMap(int x, int y)
{
  return FTerrMap[y][x];
}

void TMap::SetTerrMap(int x, int y, Byte Value)
{
  FTerrMap[y][x] = Value;
}

void TMap::SetItemMap(int x, int y, Byte Value)
{
  FItemMap[y][x] = Value;
}

void TMap::SetRole_X(int Value)
{
  FRole_X = Value;
}

void TMap::SetRole_Y(int Value)
{
  FRole_Y = Value;
}


