//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#define tileunitCPP
#include "tileunit.h"
#include "util.h"

#include <typeinfo>

const int TILE_BITS_NUM_X = 10; // ϮwCCϤƥ

//---------------------------------------------------------------------------
TTiles::TTiles()
{
  FTileNum = 0;
  FBits = new Graphics::TBitmap;
}

TTiles::~TTiles()
{
  delete FBits;
}

void TTiles::LoadFromFile(AnsiString Filename)
{
  TFileStream* fs;
  TReader* reader;

  fs = new TFileStream(Filename, fmOpenRead);
  reader = new TReader(fs, 2048);
  try {
    CheckSignature(reader, SIG_MYFILE); // ˬdY
    CheckSignature(reader, typeid(*this).name()); // ˬdY

    // FĲo property  SetXXX method
    TileNum = reader->ReadInteger();
    reader->FlushBuffer();

    // oOk, ] TTiles ڥSwqݩ, NOMwӦpŪݩ
    LoadAttrs(reader);

    FBits->LoadFromStream(fs); // ŪJsҦϤϧ

    // ˬdϧΪeŤŦXз
    if (FBits->Width < TILE_WIDTH * TILE_BITS_NUM_X)
      throw Exception("Width of tile bitmap is invalid");

    if (FBits->Height / TILE_HEIGHT * TILE_BITS_NUM_X < FTileNum)
      throw Exception("Height of tile bitmap is invalid"); // too small tiles
  } __finally {
    delete reader;
    delete fs;
  }
}

void TTiles::SaveToFile(AnsiString Filename)
{
  TFileStream* fs;
  TWriter* writer;

  fs = new TFileStream(Filename, fmCreate | fmOpenWrite);
  writer = new TWriter(fs, 2048);
  try {
    writer->WriteString(SIG_MYFILE);
    writer->WriteString(typeid(*this).name());

    writer->WriteInteger(FTileNum);
    writer->FlushBuffer();

    WriteAttrs(writer);

    FBits->SaveToStream(fs);
  } __finally {
    delete writer;
    delete fs;
  }
}

void TTiles::ImportPicture(AnsiString Filename)
{
  FBits->LoadFromFile(Filename);
  TileNum = FBits->Height / TILE_HEIGHT * TILE_BITS_NUM_X;

  // N bitmap һݪjp
  FBits->Width = TILE_WIDTH * TILE_BITS_NUM_X;
  FBits->Height = TileNum / TILE_BITS_NUM_X * TILE_HEIGHT;
}

void TTiles::SetTileNum(int Value)
{
  if (Value != FTileNum)
    FTileNum = Value;
}

int TTiles::GetTilePos_Left(int index)
{
  return (index % TILE_BITS_NUM_X) * TILE_WIDTH;
}

int TTiles::GetTilePos_Top(int index)
{
  return (index / TILE_BITS_NUM_X) * TILE_HEIGHT;
}

int TTiles::GetTileNumPerRow()
{
  return TILE_BITS_NUM_X;
}

int TTiles::GetTileRowCount()
{
  return FBits->Height / TILE_HEIGHT;
}

// implementation of TTerrTiles class

void TTerrTiles::SetTileNum(int Value)
{
  TTiles::SetTileNum(Value);
  FAttrs.resize(TileNum);
}

void TTerrTiles::LoadAttrs(TReader* filer)
{
  for (int i = 0; i < TileNum; i++)
    for (TTerrAttrElement x = taCanPass; x <= taTarget; x = (TTerrAttrElement)(x + 1)) {
      bool b = filer->ReadBoolean();
      if (b) FAttrs[i] = FAttrs[i] << x;
    }

  filer->FlushBuffer();
}

void TTerrTiles::WriteAttrs(TWriter* filer)
{
  for (int i = 0; i < TileNum; i++)
    for (TTerrAttrElement x = taCanPass; x <= taTarget; x = (TTerrAttrElement)(x + 1))
      filer->WriteBoolean(FAttrs[i].Contains(x));

  filer->FlushBuffer();
}

TTerrAttr& TTerrTiles::GetAttrs(int index)
{
  return FAttrs[index];
}

void TTerrTiles::SetAttrs(int index, const TTerrAttr& Value)
{
  FAttrs[index] = Value;
}

// implementation of TItemTiles class

void TItemTiles::SetTileNum(int Value)
{
  TTiles::SetTileNum(Value);
  FAttrs.resize(TileNum);
}

void TItemTiles::LoadAttrs(TReader* filer)
{
  for (int i = 0; i < TileNum; i++)
    for (TItemAttrElement x = iaCanMove; x <= iaSource; x = (TItemAttrElement)(x + 1))
      if (filer->ReadBoolean()) FAttrs[i] << x;

  filer->FlushBuffer();
}

void TItemTiles::WriteAttrs(TWriter* filer)
{
  for (int i = 0; i < TileNum; i++)
    for (TItemAttrElement x = iaCanMove; x <= iaSource; x = (TItemAttrElement)(x + 1))
      filer->WriteBoolean(FAttrs[i].Contains(x));

  filer->FlushBuffer();
}

TItemAttr& TItemTiles::GetAttrs(int index)
{
  return FAttrs[index];
}

void TItemTiles::SetAttrs(int index, const TItemAttr& Value)
{
  FAttrs[index] = Value;
}

#pragma package(smart_init)
